;; jitai.el is free software for Mule.
;; You can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; jitai.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;; Author MUGITANI Kunio (mugi@zinbun.kyoto-u.ac.jp)
;; Last modified at 20/11/1995

;; Comments
;; jitai.el provides the folowing character translating features.
;;   1. Function "toseiji" translates Japanese modern character to 
;;      Japanese classic character.
;;   2. Function "tosinji" translates Japanese classic character to
;;      Japanese modern character.
;;   3. Function "jistogb" translates JIS0208.1983 character to GB2312 
;;      character.
;;   4. Function "jistobig5" translates JIS0208.1980 character to BIG5
;;      character.
;;   5. Function "jishtobig5" translates JIS0208.1980 and JIS0212 character
;;      to BIG5 character.
;;   6. Function "gbtojis" translates GB2312 character to JIS0208.1980
;;      character.
;;   7. Function "gbtojish" translates GB2312 character to JIS0208.1980 and
;;      JIS0212 character.
;;   8. Function "gbtobig5" translates GB2312 ccharacter to BIG5 character.
;;   9. Function "big5tojis" translates BIG5 character to JIS0208.1980
;;      character.
;;  10. Function "big5tojish" translates BIG5 charactert to JIS0208.1980 and
;;      JIS0212 character.
;;  11. Function "big5togb" translates BIG5 character to GB2312 character.
;;  
;; Usage
;;   First, you put this jitai.el and *****-table.el files into
;;   an appropriate directory (ex.~/elisp) , and add next few lines to your
;;   .emacs file and save it.
;;      (setq load-path (cons (expand-file-name "~/elisp") load-path))
;;      (autoload 'toseiji "jitai" t t)
;;      (autoload 'tosinji "jitai" t t)
;;      (autoload 'jistogb "jitai" t t)
;;      (autoload 'jistobig5 "jitai" t t)
;;      (autoload 'jishtobig5 "jitai" t t)
;;      (autoload 'gbtojis "jitai" t t)
;;      (autoload 'gbtojish "jitai" t t)
;;      (autoload 'gbtobig5 "jitai" t t)
;;      (autoload 'big5togb "jitai" t t)
;;      (autoload 'big5tojis "jitai" t t)
;;      (autoload 'big5tojish "jitai" t t)
;;   In the buffer, you set region and type M-x Function (ex.M-x toseiji).
;;   If the source character has more then two candidates, then you will see
;;   the candidate list in the mini buffer. In this case, your choice is
;;      1. type the number of apropriate candidate. This will replace one 
;;         character with the candidate.
;;      2. type space bar. This will replace one character with the number 1
;;         candidate.
;;      3. type '!'. This will replace all targets with the number 1 candidate.
;;      4. type 'delete key'. This will skip replacing.
;;   If you feel annoying this query mode, you set "jitai-confirm" variable
;;   to nil or add 
;;      (setq jitai-confirm nil)
;;   to your .emacs file. 
;;   Function "gbtobig5" has another variable called "jitai-confirm-nonhanzi".
;;   If this variable is set to 't', then you can also select the candidates
;;   that code is between '0x2121' and '0x2f2e'.
;;   
;;   Compilation
;;   If you want to speed up this utility, you must to byte-compile all .el
;;   files. But bytecomp.el that emacs version 19 and mule version 2 provide 
;;   consume very very long time to compile ****-table.el files. 
;;
;;   Bugs
;;   ****tojis or ****tojish functions translate some characters to JIS0212
;;   hex code using "$B!c#3#1#2#1!d(B" style or translate to a few components
;;   of kanji character using "$B!P?M;0!Q(B" (= 0x304a JIS0212) style. If you
;;   don't want like this tyle, please replace all characters of this style
;;   in ****-table.el files with "$B"#(B" or a appropriate character.
;;   

(provide 'jitai)

(defconst jitai-j-base 176)
(defconst jitai-c-base 161)
(defconst jitai-f-base 193)
(defconst jitai-confirm nil "$B4A;z$KJ#?t8tJd$,M-$k>l9g$N3NG'$r$9$k$+$I$&$+$N%U%i%0(B")
(defconst jitai-confirm-nonhanzi nil "$B4A;z0J30$GJ#?t8tJd$,M-$k>l9g$N3NG'$r$9$k$+$I$&$+$N%U%i%0(B")
(defvar jitai-already-list nil)

(defun toseiji ()
  "$B#J#I#SBh0l!"BhFs?e=`$NHO0OFb$G@5;zBN$KJQ49$9$k!#(B"
  (require 'toseiji-table)
  (interactive)
  (let ((convert-table toseiji-table)
	(base-offset jitai-j-base)
	(leading-char (eval lc-jp))
	(char-category "\\cC")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jitai-convert)))

(defun tosinji ()
  "$B#J#I#SBh0l!"BhFs?e=`$NHO0OFb$G?7;zBN$KJQ49$9$k!#(B"
  (require 'tosinji-table)
  (interactive)
  (let ((convert-table tosinji-table)
	(base-offset jitai-j-base)
	(leading-char (eval lc-jp))
	(char-category "\\cC")
	(jitai-confirm nil))
    (jitai-convert)))

(defun gbtojis ()
  "$B#G#B4A;z$r#J#I#S4A;z$KJQ49$9$k!#(B"
  (require 'gbtojis-table)
  (interactive)
  (let ((convert-table gbtojis-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-cn))
	(char-category "\\cc")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jitai-convert)))

(defun gbtojish ()
  "$B#G#B4A;z$r#J#I#SJd=u4A;z$b;H$C$F#J#I#S4A;z$KJQ49$9$k!#(B"
  (require 'gbtojish-table)
  (interactive)
  (let ((convert-table gbtojish-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-cn))
	(char-category "\\cc")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jitai-convert)))

(defun jistogb ()
  "$B#J#I#S4A;z$r#G#B4A;z$KJQ49$9$k!#(B"
  (require 'jistogb-table)
  (interactive)
  (let ((convert-table jistogb-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-jp))
	(char-category "\\cj")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jitai-convert)))

(defun jistobig5 ()
  "$B#J#I#S4A;z$r#B#I#G#54A;z$KJQ49$9$k!#(B"
  (require 'jistobig5-table)
  (interactive)
  (let ((convert-table jistobig5-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-jp))
	(char-category "\\cj")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jitai-convert)))

(defun jishtobig5 ()
  "$B#J#I#S!"#J#I#SJd=u4A;z$r#B#I#G#54A;z$KJQ49$9$k!#(B"
  (require 'jishtobig5-table)
  (interactive)
  (let ((convert-table jishtobig5-table)
	(base-offset jitai-j-base)
	(leading-char (eval lc-jp2))
	(char-category "\\cj")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jistobig5)
    (jitai-convert)))

(defun gbtobig5 ()
  "$B#G#B4A;z$r#B#I#G#54A;z$KJQ49$9$k!#(B"
  (require 'gbtobig5-table)
  (interactive)
  (let ((convert-table gbtobig5-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-cn))
	(char-category "\\cc")
	(jitai-confirm (yes-or-no-p "confirm candidates ? ")))
    (jitai-convert)))

(defun big5tojis ()
  "$B#B#I#G#54A;z$r#J#I#SBh0lBhFs?e=`4A;z$KJQ49$9$k!#(B"
  (require 'big5-1tojis-table)
  (require 'big5-2tojis-table)
  (interactive)
  (big5-1tojis)
  (big5-2tojis))

(defun big5-1tojis ()
  (let ((convert-table big5-1tojis-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-big5-1))
	(char-category "[$(0!!(B-$(0_~(B]")
	(jitai-confirm nil))
    (jitai-convert)))

(defun big5-2tojis ()
  (let ((convert-table big5-2tojis-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-big5-2))
	(char-category "[$(1!!(B-$(1r~(B]")
	(jitai-confirm nil))
    (jitai-convert)))

(defun big5tojish ()
  "$B#B#I#G#54A;z$r#J#I#SBh#1!"Bh#2?e=`!"Jd=u4A;z$KJQ49$9$k!#(B"
  (require 'big5-1tojish-table)
  (require 'big5-2tojish-table)
  (interactive)
  (big5-1tojish)
  (big5-2tojish))

(defun big5-1tojish ()
  (let ((convert-table big5-1tojish-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-big5-1))
	(char-category "[$(0!!(B-$(0_~(B]")
	(jitai-confirm nil))
    (jitai-convert)))

(defun big5-2tojish ()
  (let ((convert-table big5-2tojish-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-big5-2))
	(char-category "[$(1!!(B-$(1r~(B]")
	(jitai-confirm nil))
    (jitai-convert)))

(defun big5togb ()
  "$B#B#I#G#54A;z$r#G#B4A;z$KJQ49$9$k!#(B"
  (require 'big5-1togb-table)
  (require 'big5-2togb-table)
  (interactive)
  (big5-1togb)
  (big5-2togb))

(defun big5-1togb ()
  (let ((convert-table big5-1togb-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-big5-1))
	(char-category "[$(0!!(B-$(0_~(B]")
	(jitai-confirm nil))
    (jitai-convert)))

(defun big5-2togb ()
  (let ((convert-table big5-2togb-table)
	(base-offset jitai-c-base)
	(leading-char (eval lc-big5-2))
	(char-category "[$(1!!(B-$(1r~(B]")
	(jitai-confirm nil))
    (jitai-convert)))

(defun jitai-convert()
  "Convert routin."
  (setq start (current-time-string))
  (let ((buffer-read-only nil))
    (save-excursion
      (save-restriction
	(setq jitai-already-list nil)
	(narrow-to-region (region-beginning) (region-end))
	(goto-char (point-min))
	(while (not (eobp))
	  (if (re-search-forward char-category nil t)
	      (progn 
		(setq char (char-before (point))
		      l-char (char-component char 0))
		(if (= l-char leading-char)
		    (progn
		      (setq msb (char-component char 1)
			    lsb (char-component char 2))
		      (if (setq seiji (aref convert-table (+ (* (if (< msb base-offset) 0 (- msb base-offset)) 94) (- lsb 161))))
			  (progn
			    (if (listp seiji)
				(if (< msb 176) 
				    (if jitai-confirm-nonhanzi
					(select (char-to-string char) seiji))
				  (if jitai-confirm
				      (select (char-to-string char) seiji)))
			      (progn
				(if (listp seiji)
				    (setq seiji (car seiji)))
				(delete-char -1)
				(insert seiji))))))))
	    (goto-char (point-max))))))))

(defun select (target strings)
 "" 
 (if (not (member target jitai-already-list))
     (let* ((c 0)
	    (l (length strings))
	    (selects (make-vector l ""))
	    (current-point (make-marker))
	    num)
       (setq current-point (point))
       (while (< c l)
	 (aset selects c (nth c strings))
	 (setq c (1+ c)))
       (forward-char -1)
       (while (not (eobp))
	 (if (search-forward target nil t)
	     (progn
	       (setq list-string ""
		     c 0)
	       (while (< c l)
		 (setq list-string (concat list-string (format "%d %s " (1+ c) (aref selects c)))
		       c (1+ c)))
	       (message list-string)
	       (setq num nil)
	       (while (not num)
		 (setq num (read-char)) 
		 (cond ((= num 32)
			(setq num 0))
		       ((= num 127))
		       ((= num ?!))
 		       ((and (>= num 49) (< num (+ l 49))
			     (setq num (- num 49))))
		       (t 
			(beep)
			(setq num nil)
			(message list-string))))
	       (if (= num 127)
		   nil
		 (if (= num ?!)
		     (progn 
		       (forward-char -1)
		       (while (search-forward target nil t)
			 (replace-match (aref selects 0)))
		       (goto-char (point-max)))
		   (delete-char -1) 
		   (insert (aref selects num))
		   (setq tc (aref selects num)
			 c num)
		   (while (> c 0)
		     (aset selects c (aref selects (1- c)))
		     (setq c (1- c)))
		   (aset selects 0 tc))))
	   (goto-char (point-max))))
       (goto-char current-point)
       (setq jitai-already-list (append jitai-already-list (list target))))))
